// ---------------------------- Includes -------------------------- //
#include <stdlib.h>         // C++ standard library
#include <stdexcept>
#include <stack>            // We use the standard C++ stack implementation to create model matrix stacks
#include <unistd.h>         // Threading
#include <stdio.h>          // Input/Output
#include <GLEW/glew.h>      // OpenGL Extension Wrangler -
#include <GLFW/glfw3.h>     // Windows and input
#include <glm/glm.hpp>      // OpenGL math library
#include <glm/gtc/constants.hpp>
#include "shader_util.h"    // Utility methods to keep this file a bit shorter.

/**
 * The subtask 1. (make sphere smoother) and 1.1 (send correct uniforms) are in the application code.
 * Rest are in the fragment shader.
 */

// --------------- Forward declarations ------------- //
GLuint createSphere();

// --- Load the shaders declared in glsl files in the project folder ---//
shader_prog shader("shaders/planet.vert.glsl", "shaders/planet.frag.glsl");

//We store the VAO for the planet also the face count.
GLuint planetVAO;
int planetFaceCount = 0;


/**
 * Use this method to initialize the planet
*/
void initPlanet() {
    planetVAO = createSphere();
}

/**
 * Draws the planet.
 * Applices transformations, makes the draw call.
 */
void drawPlanet() {
    std::stack<glm::mat4> ms;
    ms.push(glm::mat4(1.0));
    glm::mat4 model;

    ms.push(ms.top());
        ms.top() = glm::rotate(ms.top(), glm::radians(90.0f) * float(glfwGetTime()) / 20.0f, glm::vec3(0.0, 1.0, 0.0));
        ms.top() = glm::scale(ms.top(), glm::vec3(7.0, 7.0, 7.0));
        shader.uniformMatrix4fv("modelMatrix", ms.top());
        glBindVertexArray(planetVAO);
        glDrawElements(GL_TRIANGLES, planetFaceCount * 3, GL_UNSIGNED_SHORT, 0);
    ms.pop();
}


/**
 * This function will create the sphere geometry.
 * It uses the same algorithm as in Three.js SphereGeometry class:
 * http://threejs.org/docs/#Reference/Extras.Geometries/SphereGeometry
 *
 * --Task 1--
 * Take a look at this algorithm, and make the sphere be more smoother.
 */
GLuint createSphere() {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

    float radius = 1.0;

	int widthSegments = 8;
	int heightSegments = 6;

	float phiStart = 0.0;
	float phiLength = glm::pi<float>() * 2.0;

	float thetaStart = 0.0;
	float thetaLength = glm::pi<float>();

	int x, y;
	float u, v;
	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
	std::vector<glm::vec3> normals = std::vector<glm::vec3>();
	std::vector<glm::u16vec3> faces = std::vector<glm::u16vec3>();
	std::vector<unsigned short> indices = std::vector<unsigned short>();

	for(y = 0; y <= heightSegments; y++) {
		for(x = 0; x <= widthSegments; x++) {
			u = (float)x / widthSegments;
			v = (float)y / heightSegments;

			glm::vec3 vertex = glm::vec3(
                -radius * glm::cos(phiStart + u * phiLength) * glm::sin(thetaStart + v * thetaLength),
                radius * glm::cos(thetaStart + v * thetaLength),
                radius * glm::sin(phiStart + u * phiLength) * glm::sin(thetaStart + v * thetaLength)
            );

			vertices.push_back(vertex);
			normals.push_back(glm::normalize(vertex));
		}
	}

	int vIndex1, vIndex2, vIndex3, vIndex4;

	for(y = 0; y < heightSegments; y++) {
		for(x = 0; x < widthSegments; x++) {

            vIndex1 = (widthSegments+1) * y + x+1;
            vIndex2 = (widthSegments+1) * y + x;
            vIndex3 = (widthSegments+1) * (y+1) + x;
            vIndex4 = (widthSegments+1) * (y+1) + x + 1;

			glm::vec3 v1 = vertices.at(vIndex1);
			glm::vec3 v2 = vertices.at(vIndex2);
			glm::vec3 v3 = vertices.at(vIndex3);
			glm::vec3 v4 = vertices.at(vIndex4);

			if (glm::abs(v1.y) == radius) {
                faces.push_back(glm::u16vec3(vIndex1, vIndex3, vIndex4));
			} else if(glm::abs(v3.y) == radius) {
                faces.push_back(glm::u16vec3(vIndex1, vIndex2, vIndex3));
			} else {
                faces.push_back(glm::u16vec3(vIndex1, vIndex2, vIndex4));
                faces.push_back(glm::u16vec3(vIndex2, vIndex3, vIndex4));

				planetFaceCount += 3;
			}
			planetFaceCount += 3;
		}
	}

    shader.attributeVectorVec3("position", vertices);
    shader.attributeVectorVec3("normal", normals);

    GLuint vboHandle;
    glGenBuffers(1, &vboHandle);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboHandle);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort)*faces.size()*3, &faces[0], GL_STATIC_DRAW);

    glBindVertexArray(0);

    return vaoHandle;
}

// ---------------------------- Input -------------------------- //
// This method is called when keyboard event happens.
// Sets GLFW window should close flag to true, when escape key is pressed.
static void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
    if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
        glfwSetWindowShouldClose(window, GL_TRUE);
    }
}

// ---------------------------- Main -------------------------- //
int main(int argc, char *argv[]) {
    GLFWwindow *win;
    if (!glfwInit()) {
        exit (EXIT_FAILURE);
    }

    float screenWidth = 800;
    float screenHeight = 450;
    win = glfwCreateWindow(screenWidth, screenHeight, "Perlin Planet", NULL, NULL);

    if (!win) {
        glfwTerminate();
        exit(EXIT_FAILURE);
    }

    glfwMakeContextCurrent(win);

    //Some extensions are marked as experimental. To use the latest
    //version of OpenGL supported, we have to set this flag to true.
    glewExperimental = GL_TRUE;
    GLenum status = glewInit();

    if(status != GLEW_OK) {
        fprintf(stderr, "Error: %s\n", glewGetErrorString(status));
    }

    const GLubyte* renderer = glGetString (GL_RENDERER); // get renderer string
    const GLubyte* version = glGetString (GL_VERSION); // version as a string
    printf ("Renderer: %s\n", renderer);
    printf ("OpenGL version supported %s\n", version);

    glfwSetKeyCallback(win, key_callback);

    shader.use();

    glm::vec3 lightPosition = glm::vec3(50.0, 40.0, 20.0);
    glm::vec3 colorWater = glm::vec3(0.0, 0.68, 0.93);
    glm::vec3 colorAtm = glm::vec3(0.4, 0.84, 0.93);
    glm::vec3 color1 = glm::vec3(0.1, 0.48, 0.20);
    glm::vec3 color2 = glm::vec3(0.0, 0.35, 0.15);
    glm::vec3 color3 = glm::vec3(0.31, 0.25, 0.2);

    /**
     * -- Task 1.1. -- Pass planet colors to the shader (You could use the ones defined above or find yours)
     * The light position you can send like this:
     *     shader.uniform3vec("lightPosition", lightPosition);
     *
     * Make sure, that you are using those variables in the shader, otherwise they will get optimized out.
     */

    glm::vec3 viewer = glm::vec3(0.0, 0.0, 15.0);
    glm::mat4 perspective = glm::perspective(glm::radians(80.0f), screenWidth / screenHeight, 0.1f, 100.f);
    glm::mat4 view = glm::lookAt(
        viewer, //Position
        glm::vec3(0.0, 0.0, 0.0),  //LookAt
        glm::vec3(0.0, 1.0, 0.0)   //Up
    );

    initPlanet();

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    while (!glfwWindowShouldClose(win)) {
        shader.uniformMatrix4fv("projectionMatrix", perspective);
        shader.uniformMatrix4fv("viewMatrix", view);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        drawPlanet();

        glfwSwapBuffers(win);
        usleep(10000);
        glfwPollEvents();
    }

    glfwTerminate();
    exit(EXIT_SUCCESS);
    return 0;
}
